class CannotBeNoneError(Exception):
    pass
    
class ColumnDefaultMismatchError(Exception):
    pass

class ColumnNotFoundError(Exception):
    pass

class ColumnNullableError(Exception):
    pass

class ColumnNullableError(Exception):
    pass

class ColumnTypeMismatchError(Exception):
    pass

class InvalidColumnTypeError(Exception):
    pass

class InvalidValueTypeError(Exception):
    pass

class TableNotFoundError(Exception):
    pass

